#!/usr/bin/perl
### Read blast hit and determine protein

use strict;
use warnings;
use Bio::DB::Fasta;

my $PAR_EXTENDPERCENT = 20;		# Extend X percent up- and downstream for each hit
my $PAR_MAXSHRINKEDPERCENT = 80;	# ORF with detected start and stop codon may not be shorter than X% of original protein

my $file = pop @ARGV;		
our %known_stops;			# hash for known ORF stops
our %trg_length;			# target Fasta length
our $db;				# BioPerl Fasta
our $OUT_OF_FOUNDS_FLAG = 0;		# 1 if last operation was out of bounds, 2 if last operation lead to an unacceptable value			
our $debug = 0;				# Debug mode with additional output
our $addcounter = 0;			# Addcounter to prevent duplicated protein IDs

# http://www.sci.sdsu.edu/~smaloy/MicrobialGenetics/topics/in-vitro-genetics/codon-usage.html
our %amino = (
        'TTT' => 'F',        
        'TTC' => 'F',
        'TTA' => 'L',
        'TTG' => 'L',
        'CTT' => 'L',
        'CTC' => 'L',
        'CTA' => 'L',
        'CTG' => 'L',
        'ATT' => 'I',
        'ATC' => 'I',
        'ATA' => 'I',
        'ATG' => 'M',
        'GTT' => 'V',
        'GTC' => 'V',
        'GTA' => 'V',
        'GTG' => 'V',
        'TCT' => 'S',
        'TCC' => 'S',
        'TCA' => 'S',
        'TCG' => 'S',
        'CCT' => 'P',
        'CCC' => 'P',
        'CCA' => 'P',
        'CCG' => 'P',
        'ACT' => 'T',
        'ACC' => 'T',
        'ACA' => 'T',
        'ACG' => 'T',
        'GCT' => 'A',
        'GCC' => 'A',
        'GCA' => 'A',
        'GCG' => 'A',
        'TAT' => 'Y',
        'TAC' => 'Y',
        'TAA' => '*',
        'TAG' => '*',
        'CAT' => 'H',
        'CAC' => 'H',
        'CAA' => 'Q', 
	'CAG' => 'Q', 
        'AAT' => 'N',
        'AAC' => 'N',
        'AAA' => 'K',
        'AAG' => 'K',
        'GAT' => 'D',
        'GAC' => 'D',
        'GAA' => 'E',
        'GAG' => 'E',
        'TGT' => 'C',
        'TGC' => 'C',
        'TGA' => '*',
        'TGG' => 'W',
        'CGT' => 'R',
        'CGC' => 'R',
        'CGA' => 'R',
        'CGG' => 'R',
        'AGT' => 'S',
        'AGC' => 'S',
        'AGA' => 'R',
        'AGG' => 'R',
        'GGT' => 'G',
        'GGC' => 'G',
        'GGA' => 'G',
        'GGG' => 'G'
    );
our %amino_start = (		# Number say, it can be a start codon, but only take it if there is no alternative within [NUMBER] codons
        'TTT' => 'F',        
        'TTC' => 'F',
        'TTA' => 'L',
        'TTG' => '10',		# 3%
        'CTT' => 'L',
        'CTC' => 'L',
        'CTA' => 'L',
        'CTG' => '16',		# <1%
        'ATT' => '16',		# <1%
        'ATC' => 'I',
        'ATA' => 'I',
        'ATG' => '0',		# 83%
        'GTT' => 'V',
        'GTC' => 'V',
        'GTA' => 'V',
        'GTG' => '5',		# 14%
        'TCT' => 'S',
        'TCC' => 'S',
        'TCA' => 'S',
        'TCG' => 'S',
        'CCT' => 'P',
        'CCC' => 'P',
        'CCA' => 'P',
        'CCG' => 'P',
        'ACT' => 'T',
        'ACC' => 'T',
        'ACA' => 'T',
        'ACG' => 'T',
        'GCT' => 'A',
        'GCC' => 'A',
        'GCA' => 'A',
        'GCG' => 'A',
        'TAT' => 'Y',
        'TAC' => 'Y',
        'TAA' => '*',
        'TAG' => '*',
        'CAT' => 'H',
        'CAC' => 'H',
        'CAA' => 'Q', 
	'CAG' => 'Q', 
        'AAT' => 'N',
        'AAC' => 'N',
        'AAA' => 'K',
        'AAG' => 'K',
        'GAT' => 'D',
        'GAC' => 'D',
        'GAA' => 'E',
        'GAG' => 'E',
        'TGT' => 'C',
        'TGC' => 'C',
        'TGA' => '*',
        'TGG' => 'W',
        'CGT' => 'R',
        'CGC' => 'R',
        'CGA' => 'R',
        'CGG' => 'R',
        'AGT' => 'S',
        'AGC' => 'S',
        'AGA' => 'R',
        'AGG' => 'R',
        'GGT' => 'G',
        'GGC' => 'G',
        'GGA' => 'G',
        'GGG' => 'G'
    );

# Open reblast jobs
open(JOBS,"<$file") || die ("Could not open file '$file': $!");
while (my $job = <JOBS>){
	chomp $job;
	my ($group, $name, $genome) = split(/\t+/,$job);

	# Get protein lengths
	my %src_length = &fastacount2hash("$group.faa");
	%trg_length = &fastacount2hash($genome);
	$db = Bio::DB::Fasta->new($genome,-reindex => 0);

	# Read GFF
	%known_stops = ();
	my $gff = $genome;
	$gff =~ s/\.fna/.gff/;
	open(GFF,"<$gff") || die("$!");
	while(<GFF>) {
		my @data = split(/\s+/);
		if ($_ =~ /^#/ || !defined($data[6]) || ($data[6] ne "+" && $data[6] ne "-") || $data[2] ne "CDS") {next;}
		# NC_012438.1	RefSeq	CDS	5241	5441	.	-	
		my @given = ($data[6],$data[3],$data[4],$data[0]);
		&check_for_known_stops(@given);
	}
	close(GFF);

	# Open output GFF & FAA
	open(GFF,">>$name.add.gff") || die ("Could not open file '$name.add.gff': $!");
	open(FAA,">>$name.add.faa") || die ("Could not open file '$name.add.faa': $!");

	# Open blast output
	open(BLA,"<$group.$name.bla") || die ("Could not open file '$group.$name.bla': $!");
	while (my $bla = <BLA>) {
		chomp $bla;
		my ($query_id,$subject_id,$identity,$alignment_length,$mismatches,$openings,$query_start,$query_end,$subject_start,$subject_end,$evalue,$bitscore) = split(/\s+/,$bla);				
		unless (defined($src_length{$query_id})) {die("Error length of query sequence '$query_id' not known!");}

		print "Checking Protein group $group -> Target: $name $subject_start..$subject_end from reference $query_start..$query_end\n";

		# Extend hit downstream
		my @original_hit = &get_coords($subject_start,$subject_end,$subject_id);
		my $downstream_difference = $src_length{$query_id}-$query_end;
		my $add_downstream = 3+int(($PAR_EXTENDPERCENT/100*$src_length{$query_id}+$downstream_difference)*3);		# full length + some extra for a more downstream stop

		if ($debug) {print "BLAST match:\t".&nuc2amino(&cut_seq(@original_hit))."\n";}
		my @downstream_extended_hit = &add_downstream(@original_hit,$add_downstream);
		if ($OUT_OF_FOUNDS_FLAG) {next;}
		if ($debug) {print "DS extended:\t".&nuc2amino(&cut_seq(@downstream_extended_hit))."\n";}

#		print "Original hit: $query_start,$query_end -> $subject_start,$subject_end\n\t".join(", ",@original_hit)."\n\t".join(", ",@downstream_extended_hit)."\n";

		# Look for a matching stop codon, begin search at hit start
		my @stop_codoned_hit = &find_stop_codon(@downstream_extended_hit);

#		print join(", ",@original_hit)."\n";
#		print join(", ",@downstream_extended_hit)."\n";
		unless (defined($stop_codoned_hit[0])) {
			if ($debug) {print "Note: No stop codon found.\n";}
			next;
		}				# No hit found

		if (&check_for_known_stops(@stop_codoned_hit)) {
			if ($debug) {print "Note: This ORF position is allready in use.\n";}
			next;
		}

		if ($debug) {print "Stop codon:\t".&nuc2amino(&cut_seq(@stop_codoned_hit))."\n";}

#		print join(", ",@stop_codoned_hit)."\n";
		
#		my $x = &cut_seq(@stop_codoned_hit);
#		print &nuc2amino($x)."\n";

		# Extend hit upstream
		my $upstream_difference = $query_start-1;
		my $add_upstream = 3+int(($PAR_EXTENDPERCENT/100*$src_length{$query_id}+$upstream_difference)*3);		# full length + some extra for a more upstream start			
		my @upstream_extended_hit = &add_upstream(@stop_codoned_hit,$add_upstream);
		if ($OUT_OF_FOUNDS_FLAG) {next;}


		if ($debug) {print "US extended:\t".&nuc2amino(&cut_seq(@upstream_extended_hit))."\n";}
		# Look for a matching start codon, begin search somewhat downstream of blast hit
		my @start_codoned_hit = &find_start_codon(@upstream_extended_hit,$add_upstream,$upstream_difference*3);

		unless (defined($start_codoned_hit[0])) {
			if ($debug) {print "Note: No start codon found.\n";}
			next;
		}				# No hit found
		if ($debug) {print "Start codon:\t".&nuc2amino(&cut_seq(@start_codoned_hit))."\n";}

		# Check if remains are long enough
		if (abs($start_codoned_hit[1]-$start_codoned_hit[2])/3 < $PAR_MAXSHRINKEDPERCENT/100*$src_length{$query_id}) {
			if ($debug) {print "Note: Remaining ORF became to short: ORF ".int(abs($start_codoned_hit[1]-$start_codoned_hit[2])/3)."nt, ORG: ".$src_length{$query_id}."nt \n";}
			next;
		}

		my $sequence = &nuc2amino(&cut_seq(@start_codoned_hit));
		$addcounter++;

		# NCBI
		my @chr = split(/\|/,$start_codoned_hit[3]);
		my $chr = pop @chr;

		print FAA ">PO_$addcounter ".join(" ",@start_codoned_hit)."$group $name $subject_id from $query_id\n$sequence\n";
		print GFF "$chr\tProteinortho\tCDS\t$start_codoned_hit[1]\t$start_codoned_hit[2]\t.\t$start_codoned_hit[0]\t0\t";
		print GFF "Name=PO_$addcounter;ID=PO_$addcounter;Note=Hit from $query_id\n";
		# NC_000918.1	RefSeq	CDS	2117	3334	.	+	0	ID=cds1;Name=NP_212987.1;Parent=gene1;Note=EF-Tu%3B promotes GTP-dependent binding of aminoacyl-tRNA to the A-site of ribosomes during protein biosynthesis%3B when the tRNA anticodon matches the mRNA codon%2C GTP hydrolysis results%3B the inactive EF-Tu-GDP leaves the ribosome and release of GDP is promoted by elongation factor Ts%3B many prokaryotes have two copies of the gene encoding EF-Tu;Dbxref=Genbank:NP_212987.1,GeneID:1192534;gbkey=CDS;product=elongation factor Tu;protein_id=NP_212987.1;transl_table=11
	}
	close(BLA);
	close(FAA);
	close(GFF);
}
close(JOBS);

sub check_for_known_stops {
	my ($strand, $a, $b, $chr) = @_;

	# Special for NCBI GFFs
	if ($chr =~ /\|/) {	
		my @chr = split(/\|/,$chr);
		$chr = pop @chr;
	}

	# Stop position + 3 up and downstream
	my $stopU = $strand.($b-3).$chr;
	my $stop = $strand.$b.$chr;
	my $stopD = $strand.($b+3).$chr;

	if ($strand eq "-") {
		$stopU = $strand.($a+3).$chr;
		$stop = $strand.$a.$chr;
		$stopD = $strand.($a-3).$chr;
	}

	if (defined($known_stops{$stop}) || exists($known_stops{$stopU}) || exists($known_stops{$stopD})) {
#		print $stop."\n";
		return 1;
	}

#	print "Store ORF $stop\n";
	$known_stops{$stop} = 1;

	return 0;	
}

sub nuc2amino {
	my @letters = split(//,$_[0]);
	my $amino = "";
	for (my $i = 0; $i < scalar(@letters)-3; $i += 3) {
		my $codon = $letters[$i].$letters[$i+1].$letters[$i+2];
		unless (defined($amino{$codon})) {$amino{$codon} = "X"; print "Note: Codon '$codon' could not be converted. Using 'X' as pseudo amino acid.\n";}
		$amino .= $amino{$codon};
	}
	return $amino;
}

sub find_start_codon {
	my $should = pop;							# Position where we would expect the start
	my $added = pop;	$added = int(6 + $added - $added%3);		# Make sure we stay in frame
	$should += $added;

	if ($debug) {print "Added: $added Should: $should\n";}

	my ($strand, $a, $b, $chr) = @_;
	my $seq = &cut_seq(@_);
	my @letters = split(//,$seq);
	my $hit = -1;
	my %start_hits;
	for (my $i = $added; $i > 0; $i -= 3) {
		my $codon = $letters[$i].$letters[$i+1].$letters[$i+2];
		unless (defined($amino_start{$codon})) {$amino_start{$codon} = "X"; print "Note: Codon '$codon' could not be converted. Using 'X' as pseudo amino acid.\n";}
		my $amino = $amino_start{$codon};
		if ($amino eq "*") {last;}
		if ($amino =~ /[0-9]/) {
			$start_hits{$amino+abs($i-$should)/3} = $i;}		# Collect potential start positions; Formula: favorable vs. Expected position
	}
	
	foreach (sort keys %start_hits) {				# Get the best hit (Distance, score)
#		print "Score: $_ ->\tPosition: $start_hits{$_}\n";
		if ($hit == -1) {$hit = $start_hits{$_};}
		last;
	}
#	print "Use position $hit\n";

	if ($hit > 0) {
		return &add_upstream(@_,-1*$hit);
	}
	return undef;
}

sub find_stop_codon {
	my ($strand, $a, $b, $chr) = @_;
	my $seq = &cut_seq(@_);
	my @letters = split(//,$seq);
	my $hit = -1;
	for (my $i = 0; $i < scalar(@letters)-3; $i += 3) {
		my $codon = $letters[$i].$letters[$i+1].$letters[$i+2];
		unless (defined($amino{$codon})) {$amino{$codon} = "X"; print "Note: Codon '$codon' could not be converted. Using 'X' as pseudo amino acid.\n";}
		my $amino = $amino{$codon};
		if ($amino eq "*") {
			$hit = $i+2;
			last;
		}
	}

	if ($hit > 0) {
#		return &add_downstream(@_,-1*(abs($a-$b)-$hit-3));
		return &add_downstream(@_,-1*(abs($a-$b)-$hit));	# without stop codon
	}
	return undef;
}


# Strand generalized operations - codon safe, gff style positions ( a < b), negative values are okay
# Last item tells if operation lead to an out of bound exeption. Important if shrinking (negative) operations were applied.
sub get_coords {
	my ($a, $b, $chr) = @_;
	$chr =~ s/\s+.*$//;
	unless (defined($trg_length{$chr})) {
		die("Error: No length defined for target chromosome '$chr'\n");
	}

	my $strand = "+";
	if ($b < $a) {
		my $dummy = $a;
		$a = $b;
		$b = $dummy;
		$strand = "-";
	}
	
	if (($b-$a)%3 != 2) {die("Error: Given coordinates $strand $a ... $b are not codon like!");}

	return($strand, $a, $b, $chr);
}

sub cut_seq {
	my ($strand, $a, $b, $chr) = @_;
	my $seq = "";
	if 	($strand eq "+") {$seq = $db->subseq($chr,$a,$b);}
	elsif 	($strand eq "-") {$seq = $db->subseq($chr,$b,$a);}
	else {die("Strand $strand cannot be used!");}

	if (length($seq) != abs($a-$b)+1) {
		die("Error: Cut length does not match. Length is ".length($seq)." but should be ".abs($a-$b)+1);
	}
	return $seq;
}

sub add_upstream {
	my ($strand, $a, $b, $chr, $upstream) = @_;
	if ($debug) {print "Add upstream: $upstream\n";}
	$OUT_OF_FOUNDS_FLAG = 0;	
	
	# Codon stabilize upstream
	$upstream = int($upstream);
	$upstream -= $upstream%3;

	# Plus
	if ($strand eq "+") {
		$a -= $upstream;
		# Fix out of string
		if ($a < 1) {
			$a -= int($a/3)*3;			# $a is negative, so we acutually add something here
			if ($a < 1) {$a += 3;}
			$OUT_OF_FOUNDS_FLAG = 1;
		}
		if ($a > $trg_length{$chr}) {
			$a -= int($a-($trg_length{$chr})/3)*3;
			if ($a > $trg_length{$chr}) {$a -= 3;}
			$OUT_OF_FOUNDS_FLAG = 1;
		}		
	}
	elsif ($strand eq "-") {
		$b += $upstream;
		# Fix out of string
		if ($b < 1) {
			$b -= int($b/3)*3;			# $b is negative, so we acutually add something here
			if ($b < 0) {$b += 3;}
			$OUT_OF_FOUNDS_FLAG = 1;
		}
		if ($b > $trg_length{$chr}) {
			$b -= int($b-($trg_length{$chr})/3)*3;
			if ($b > $trg_length{$chr}) {$b -= 3;}
			$OUT_OF_FOUNDS_FLAG = 1;
		}
	}
	else {die("Strand $strand cannot be used!");}

	if ($b <= $a) {$OUT_OF_FOUNDS_FLAG = 2; print STDERR "Postion became $a ... $b -> not compatible with gff standard\n";}

	return ($strand, $a, $b, $chr);
}

sub add_downstream {
	my ($strand, $a, $b, $chr, $downstream) = @_;
	if ($debug) {print "Add downstream: $downstream\n";}
	$OUT_OF_FOUNDS_FLAG = 0;	
	
	# Codon stabilize upstream
	$downstream = int($downstream);
	$downstream -= $downstream%3;

	# Plus
	if ($strand eq "+") {
		$b += $downstream;
		# Fix out of string
		if ($b < 1) {				
			$b -= int($b/3)*3;			# $b is negative, so we acutually add something here
			if ($b < 1) {$b += 3;}
			$OUT_OF_FOUNDS_FLAG = 1;
		}
		if ($b > $trg_length{$chr}) {
			$b -= int($b-($trg_length{$chr})/3)*3;
			if ($b > $trg_length{$chr}) {$b -= 3;}
			$OUT_OF_FOUNDS_FLAG = 1;
		}
	}
	elsif ($strand eq "-") {
		$a -= $downstream;
		if ($a < 1) {
			$a -= int($a/3)*3;			# $a is negative, so we acutually add something here
			if ($a < 1) {
				$a += 3;
			}
			$OUT_OF_FOUNDS_FLAG = 1;
		}
		if ($a > $trg_length{$chr}) {
			$a -= int($a-($trg_length{$chr})/3)*3;
			if ($a > $trg_length{$chr}) {$a -= 3;}
			$OUT_OF_FOUNDS_FLAG = 1;
		}
	}
	else {die("Strand $strand cannot be used!");}

	if ($b <= $a) {$OUT_OF_FOUNDS_FLAG = 2; print STDERR "Postion became $a ... $b -> not compatible with gff standard\n";}

	return ($strand, $a, $b, $chr);
}


### Fasta tools
# Read a fasta File
#my %hash = fasta2hash($file);
sub fasta2hash {
	my $file = shift;
	my $cut = shift;
	if (!defined($cut)) {$cut = 1;}		# Remove \s.* from headers (blast style)
	my %hash;
	open(FILE,"<$file") || die("fasta2hash(): Could not open file: $file\n");
	my $last_hash = "";
	while(<FILE>) {
		$_ =~ s/\r/\n/g;
		chomp;
		if ($_ =~ /^>/) {
			$_ =~ s/^>//;
			if ($cut) {
				$_ =~ s/\s+.*$//s;
			}
			$hash{$_} = "";
			$last_hash = $_;
			next;
		}
		$hash{$last_hash} .= $_;
	}
	close(FILE);
	return %hash;
}

# Read a fasta file, count lengths
# Access e.g. via $results[x]->[y]
#my %hash = fasta2hash($file);
sub fastacount2hash {
	my $file = shift;
	my $cut = shift;
	if (!defined($cut)) {$cut = 1;}		# Remove \s.* from headers (blast style)
	my %hash;
	open(FILE,"<$file") || die("fastacount2hash(): Could not open file: $file\n");
	my $last_hash = "";
	while(<FILE>) {
		$_ =~ s/\r/\n/g;
		chomp;
		if ($_ =~ /^>/) {
			$_ =~ s/^>//;
			if ($cut) {
				$_ =~ s/\s+.*$//s;
			}
			$hash{$_} = 0;
			$last_hash = $_;
			next;
		}
		$hash{$last_hash} += length($_);
	}
	close(FILE);
	return %hash;
}
